/*
Copyright 2010 Harteex Productions
License: Creative Commons Attribution 3.0 Unported
http://creativecommons.org/licenses/by/3.0/
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <stdint.h>
#include <dingoo/ucos2.h>
#include <dingoo/entry.h>

#include <sml/graphics.h>
#include <sml/sprite.h>
#include <sml/display.h>
#include <sml/control.h>
#include <sml/timer.h>

#include "images.h"
#include "achievements.h"
#include "hif.h"
#include "list.h"
#include "mixer.h"
#include "sounds.h"
#include "music.h"

#include "config.h"

#define VERSION "V1.21"

#if COMPILE_TARGET != DINGOO_A320 && COMPILE_TARGET != GEMEI_760P && COMPILE_TARGET != GEMEI_A330
#error Missing or invalid compilation target. Please set target in config.h.
#endif

#if COMPILE_TARGET == GEMEI_A330
#ifndef CONTROL_POWER
#define CONTROL_POWER 1 // Doesn't exist on the Gemei, but keep this here to make things compile
#endif
#endif

#define max(x, y) (x > y ? x : y)
#define min(x, y) (x < y ? x : y)
#define SAFE_FREE(p) { if (p) {free(p); p = 0;}}

typedef struct {
	char      name[15];
	uint32_t  score;
} highscore;

void gameScreenshot();

// MAJOR DRAWING METHODS
void drawSplash();
void drawMenu();
void drawGame();
void drawError();

// HELPER DRAWING METHODS
void drawMainMenu();
void drawHud();
void drawMarker();
void drawGrid();
void drawTile(uint16_t x, uint16_t y, uint16_t tile);
void drawHighscoreInput();
void drawInputMarker();
void drawHighscore();
void drawCredits();
void drawInstructions();
void drawVolumeMeter();

// LOGIC AND LOADING
void setGameMode(uint8_t mode);
void setGameSubMode(uint8_t mode);
uint8_t getTile(int8_t x, int8_t y);
void setTile(int8_t x, int8_t y, uint8_t tile);
void doLogic();
void doInput();
void doRemoveTile();
uint8_t getNormalColorTile(uint8_t tile);
uint8_t getTileType(uint8_t tile);
uint16_t setTo(uint8_t type, uint8_t tileToSet, int8_t x, int8_t y);
void doDisappearing();
void removeAllDisappearing();
bool isGameOver();
void createLevel();
void startGame();
bool loadGame();
void unloadGame();
bool loadData();
bool saveData();

#define GAMEWIDTH  15
#define GAMEHEIGHT 10
#define TILESIZE 20

#define TILE_NONE                0
#define TILE_RED                 1
#define TILE_BLUE                2
#define TILE_YELLOW              3
#define TILE_RED_SELECTED        4
#define TILE_BLUE_SELECTED       5
#define TILE_YELLOW_SELECTED     6
#define TILE_RED_DISAPPEARING    7
#define TILE_BLUE_DISAPPEARING   8
#define TILE_YELLOW_DISAPPEARING 9

#define TYPE_NONE          0
#define TYPE_NORMAL        1
#define TYPE_DISAPPEARING  2
#define TYPE_SELECTED      3

#define GAME_MODE_SPLASH 1
#define GAME_MODE_MENU   2
#define GAME_MODE_GAME   3
#define GAME_MODE_ERROR  4

#define GAME_MODE_MENU_MAIN           1
#define GAME_MODE_MENU_FADEIN         2
#define GAME_MODE_MENU_FADEOUT        3
#define GAME_MODE_MENU_ACHIEVEMENTS   4
#define GAME_MODE_MENU_CREDITS        5
#define GAME_MODE_MENU_INSTRUCTIONS   6

#define GAME_MODE_GAME_PLAY           1
#define GAME_MODE_GAME_ANIMATION      2
#define GAME_MODE_GAME_WRITEHIGHSCORE 3
#define GAME_MODE_GAME_VIEWHIGHSCORE  4
#define GAME_MODE_GAME_END            5

uint8_t tileAnimFrames = 10;

gfx_font* gameFont               = NULL;
gfx_font* gameFontDisabled       = NULL;
gfx_font* gameFontSmall          = NULL;
gfx_texture* harteexLogo         = NULL;
gfx_texture* gameLogo            = NULL;
gfx_texture* gameBackground      = NULL;
gfx_texture* gameBackgroundRight = NULL;
sprite* tileRed                  = NULL;
sprite* tileBlue                 = NULL;
sprite* tileYellow               = NULL;
sprite* gameMarker               = NULL;
sprite* menuArrowLeft            = NULL;
sprite* menuArrowRight           = NULL;
sprite* volumeHigh               = NULL;
sprite* volumeLow                = NULL;
sprite* volumeOff                = NULL;
sprite* tileRedAnim              = NULL;
sprite* tileBlueAnim             = NULL;
sprite* tileYellowAnim           = NULL;
sprite* tileRedSelected          = NULL;
sprite* tileBlueSelected         = NULL;
sprite* tileYellowSelected       = NULL;

uint8_t     gameMode     = GAME_MODE_SPLASH;
uint8_t     gameModeSub  = GAME_MODE_GAME_PLAY;
uint8_t     targetSub    = GAME_MODE_GAME_PLAY; // Used by FADEOUT

display* gameDisplay  = NULL;
timer*   gameTimer    = NULL;
uint32_t gameTickRate = (timer_resolution / 30);
bool     gameRunning  = true;

FILE* customMusic = NULL;

uint16_t disappearingTicks = 0;

int16_t  markerX = 0;
int16_t  markerY = 0;
uint8_t  xOffset = 0;
uint8_t  yOffset = 40;
uint32_t score = 0;
int16_t  markerInputX = 0;
int16_t  markerInputY = 0;
bool     gameInProgress = false;
bool     removedAnyRed = false; // Used for achievement
bool     removedAnyBlue = false;
bool     removedAnyYellow = false;

uint8_t  showVolumeMeter = 0;
uint8_t  startAnimation = 0;

uint8_t  grid[GAMEWIDTH * GAMEHEIGHT];

char*    main_menu [] = {"Continue", "New Game", "High Scores", "Instructions", "Achievements", "Credits", "Exit"};
uint8_t  menu_length = 7;
uint8_t  menu_index = 1;
int16_t  achievementIndex = 0;
uint8_t  instructionsPage = 1;

char*    errorMsg = "Unknown error.";

uint8_t  ticksLogic = 0;
uint8_t  inputDelay = 0;
uint8_t  delayAmount = 5;
uint16_t bgRightXMax = 272;
uint16_t bgRightXMin = 191;
uint16_t bgRightX = 272;
uint8_t  inputOffsetX = 30;
uint8_t  inputOffsetY = 60;
uint8_t  inputSpaceX = 13;
uint8_t  inputSpaceY = 26;

highscore highscores[6];

uint8_t  volumeConfig = 20;

char inputChars[] = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R',
                     'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '', '', '', '', '', '', '', '', '', ' ',
                     'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r',
					 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '', '', '', '', '', '', '', '', '', ''};

char inputCurName[15];
char inputLastName[15];

uint16_t gameStarts = 0; // for an achievement
uint16_t aTimer = 0;

uint16_t screenshotCounter = 0;


inline uint32_t gfx_color_back(gfx_color inColor) {
		return ((inColor & 0xF800) << 8) | ((inColor & 0x07E0) << 5) | ((inColor & 0x001F) << 3);
}

uint32_t pow(uint32_t base, uint32_t exp)
{
	if (exp > 1)
		return base * pow(base, exp - 1);
	else
		return base;
}

void gameScreenshot() {
	char tempString[256];
	unsigned long int tempNumber = 0;
	FILE* tempFile;
	while(true) {
		sprintf(tempString, "screenshot%lu.tga", tempNumber);
		tempFile = fopen(tempString, "rb");
		if(tempFile == NULL)
			break;
		fclose(tempFile);
		tempNumber++;
	}
	gfx_tex_save_tga(tempString, gfx_render_target);
}

// MAJOR DRAWING METHODS

void drawSplash()
{
	gfx_render_target_clear(gfx_color_rgb(0xFF, 0xFF, 0xFF));
	if (harteexLogo != NULL)
		gfx_tex_draw(((gameDisplay->width - harteexLogo->width) >> 1), ((gameDisplay->height - harteexLogo->height) >> 1), harteexLogo);
}

void drawMenu()
{
	gfx_tex_draw(0, 0, gameBackground);
	gfx_tex_draw(bgRightX, 0, gameBackgroundRight);
	switch (gameModeSub)
	{
		case GAME_MODE_MENU_MAIN:
			drawMainMenu();
			break;
		case GAME_MODE_MENU_ACHIEVEMENTS:
			drawAchievementList();
			break;
		case GAME_MODE_MENU_CREDITS:
			drawCredits();
			break;
		case GAME_MODE_MENU_INSTRUCTIONS:
			drawInstructions();
			break;
	}
}

void drawGame()
{
	gfx_tex_draw(0, 0, gameBackground);
	gfx_tex_draw(bgRightX, 0, gameBackgroundRight);
	switch (gameModeSub)
	{
		case GAME_MODE_GAME_PLAY:
			drawHud();
			drawGrid();
			drawMarker();
			break;
		case GAME_MODE_GAME_ANIMATION:
			drawHud();
			drawGrid();
			break;
		case GAME_MODE_GAME_WRITEHIGHSCORE:
			drawHud();
			drawGrid();
			drawHighscoreInput();
			drawInputMarker();
			break;
		case GAME_MODE_GAME_VIEWHIGHSCORE:
			drawHighscore();
			break;
	}
}

void drawError()
{
	gfx_render_target_clear(gfx_color_rgb(0x00, 0x00, 0x00));
	gfx_font_print_center(gameDisplay->height / 2, gameFont, errorMsg);
}

// HELPER DRAWING METHODS

void drawMainMenu()
{
	uint8_t startY = 50;
	uint8_t spaceY = 25;
	gfx_font* font = gameFont;

	gfx_tex_draw(111 - gameLogo->width / 2, 16, gameLogo);

	uint8_t i;
	for (i = 0; i < menu_length; i++)
	{
		if (i == 0 && !gameInProgress)
		{
			font = gameFontDisabled;
		}
		else
		{
			font = gameFont;
		}

		gfx_font_print_center_ex(startY + spaceY * i, 222, 0, font, main_menu[i]);
	}
	sprite_draw(menuArrowLeft, 111 - 90 - (menuArrowLeft->width / 2), startY + spaceY * menu_index + 3, 0);
	sprite_draw(menuArrowRight, 111 + 90 - (menuArrowRight->width / 2), startY + spaceY * menu_index + 3, 0);

	// Draw version
	gfx_font_print(0, gameDisplay->height - gfx_font_height(gameFontSmall), gameFontSmall, VERSION);
}

void drawHud()
{
	char tempString[256];
	sprintf(tempString, "SCORE: %u", score);
	gfx_font_print(35, 3, gameFont, tempString);
}

void drawMarker()
{
	if (markerX < 0 || markerY < 0 || markerX >= GAMEWIDTH || markerY >= GAMEHEIGHT)
	{
		return;
	}

	uint16_t xPx = markerX * TILESIZE + xOffset;
	uint16_t yPx = markerY * TILESIZE + yOffset;
	sprite_draw(gameMarker, xPx + (TILESIZE / 2), yPx + (TILESIZE / 2), 0);
}

void drawGrid()
{
	int16_t x;
	int16_t y;
	for (y = 0; y < GAMEHEIGHT; y++)
	{
		for (x = 0; x < GAMEWIDTH; x++)
		{
			drawTile(x, y, grid[x + y * GAMEWIDTH]);
		}
	}
}

void drawTile(uint16_t x, uint16_t y, uint16_t tile)
{
	uint16_t xPx = x * TILESIZE + xOffset;
	uint16_t yPx = y * TILESIZE + yOffset;

	switch (tile)
	{
		case TILE_RED:
			sprite_draw(tileRed, xPx, yPx, 0);
			break;
		case TILE_BLUE:
			sprite_draw(tileBlue, xPx, yPx, 0);
			break;
		case TILE_YELLOW:
			sprite_draw(tileYellow, xPx, yPx, 0);
			break;
		case TILE_RED_SELECTED:
			sprite_draw(tileRedSelected, xPx, yPx, 0);
			break;
		case TILE_BLUE_SELECTED:
			sprite_draw(tileBlueSelected, xPx, yPx, 0);
			break;
		case TILE_YELLOW_SELECTED:
			sprite_draw(tileYellowSelected, xPx, yPx, 0);
			break;
		case TILE_RED_DISAPPEARING:
			sprite_draw(tileRedAnim, xPx, yPx, disappearingTicks < tileAnimFrames ? disappearingTicks : tileAnimFrames - 1);
			break;
		case TILE_BLUE_DISAPPEARING:
			sprite_draw(tileBlueAnim, xPx, yPx, disappearingTicks < tileAnimFrames ? disappearingTicks : tileAnimFrames - 1);
			break;
		case TILE_YELLOW_DISAPPEARING:
			sprite_draw(tileYellowAnim, xPx, yPx, disappearingTicks < tileAnimFrames ? disappearingTicks : tileAnimFrames - 1);
			break;
	}

	if (startAnimation < TILESIZE / 2)
		gfx_rect_fill_draw(xPx + startAnimation, yPx + startAnimation, TILESIZE - startAnimation * 2, TILESIZE - startAnimation * 2, gfx_color_rgb(0x73, 0xae, 0xde));
}

void drawHighscoreInput()
{
	uint16_t i;
	uint8_t x;
	uint8_t y;

	char tempString[256];
	sprintf(tempString, "Name: %s", inputCurName);
	gfx_font_print(inputOffsetX, inputOffsetY - 26, gameFont, tempString);

	for (i = 0; i < 72; i++)
	{
		x = i % 18;
		y = i / 18;

		gfx_font_print_char(inputOffsetX + x * inputSpaceX, inputOffsetY + y * inputSpaceY, gameFont, inputChars[i]);
	}

	gfx_font_print(inputOffsetX + 4 * inputSpaceX, inputOffsetY + (inputSpaceY * 4), gameFont, "Backspace");
	gfx_font_print(inputOffsetX + 15 * inputSpaceX, inputOffsetY + (inputSpaceY * 4), gameFont, "End");

	gfx_line_draw(inputOffsetX, inputOffsetY, inputOffsetX + 18 * inputSpaceX, inputOffsetY, gfx_color_rgb(0x2e, 0x44, 0x56));
	gfx_line_draw(inputOffsetX, inputOffsetY + 4 * inputSpaceY, inputOffsetX + 18 * inputSpaceX, inputOffsetY + 4 * inputSpaceY, gfx_color_rgb(0x2e, 0x44, 0x56));
}

void drawInputMarker()
{
	if (markerInputX < 0 || markerInputY < 0)
	{
		return;
	}

	uint16_t xPx = markerInputX * inputSpaceX + inputOffsetX;
	uint16_t yPx = markerInputY * inputSpaceY + inputOffsetY;
	sprite_draw(gameMarker, xPx + (inputSpaceX / 2), yPx + (inputSpaceY / 2), 0);
}

void drawHighscore()
{
	uint16_t xStart = inputOffsetX;
	uint16_t yStart = 40;
	uint8_t spaceY = 26;
	char tempString[256];
	uint8_t i;
	for (i = 0; i < 6; i++)
	{
		sprintf(tempString, "%-20s%4u", highscores[i].name, highscores[i].score);
		gfx_font_print(xStart, yStart + i * spaceY, gameFont, tempString);
	}

	sprintf(tempString, "Press A to %s", gameInProgress ? "continue the game" : "start a new game");
	gfx_font_print(xStart + 10, 210, gameFontSmall, tempString);

	if (!gameInProgress)
	{
		sprintf(tempString, "LAST GAME SCORE: %u", score);
		gfx_font_print(35, 3, gameFont, tempString);
	}
}

void drawAchievementList()
{
	uint8_t i;
	sprite* image;
	uint16_t xStart = 30;
	uint16_t yStart = 30;
	uint16_t ySpace = 20;

	for (i = 0; i < 8; i++)
	{
		if (isAchievementUnlocked(i))
			image = achievementUnlocked;
		else
			image = achievementLocked;

		char* name = achievementNames[i];
		char* desc = achievementDescriptions[i];

		if (achievementSecret[i] && !isAchievementUnlocked(i))
		{
			name = "Secret";
			desc = "Continue playing to\nunlock";
		}

		sprite_draw(image, xStart, yStart + i * ySpace, 0);
		gfx_font_print(xStart + image->width + 5, yStart + i * ySpace - 2, gameFont, name);
		if (i == achievementIndex)
		{
			sprite_draw(menuArrowLeft, 12, yStart + i * ySpace + 1, 0);
			gfx_font_print(15, 196, gameFontSmall, desc);
		}
	}
}

void drawCredits()
{
	gfx_font_print_center_ex(10, 222, 0, gameFont, "Programming");
	gfx_font_print_center_ex(30, 222, 0, gameFontSmall, "Harteex");
	gfx_font_print_center_ex(50, 222, 0, gameFont, "Graphics");
	gfx_font_print_center_ex(70, 222, 0, gameFontSmall, "From Dat Nguyen etc");
	gfx_font_print_center_ex(90, 222, 0, gameFont, "Music");
	gfx_font_print_center_ex(110, 222, 0, gameFontSmall, "Daniel Kvarfordt");
	gfx_font_print_center_ex(130, 222, 0, gameFont, "Thanks to");
	gfx_font_print_center_ex(150, 222, 0, gameFontSmall, "Flatmush");
	gfx_font_print_center_ex(170, 222, 0, gameFontSmall, "Alekmaul");
	gfx_font_print_center_ex(190, 222, 0, gameFontSmall, "Share-TechMono font by");
	gfx_font_print_center_ex(210, 222, 0, gameFontSmall, "Ralph Oliver du Carrois");
}

void drawInstructions()
{
	char tempString[256];
	uint16_t xStart = 15;

	if (instructionsPage == 1)
	{
		gfx_font_print(xStart, 20, gameFont, "Goal");
		gfx_font_print(xStart, 40 + 16 * 0, gameFontSmall, "The goal in SameGoo is\nto get a high score.");
		gfx_font_print(xStart, 40 + 16 * 3, gameFontSmall, "Selecting an object will\nclear that object and\nall connected objects of\nthe same color.");
		gfx_font_print(xStart, 40 + 16 * 8, gameFontSmall, "There must be at least\ntwo connected objects.");
	}
	else if (instructionsPage == 2)
	{
		gfx_font_print(xStart, 20, gameFont, "Goal");
		gfx_font_print(xStart, 40 + 16 * 0, gameFontSmall, "The more objects cleared\nat once, the higher\nscore you will get.");
		gfx_font_print(xStart, 40 + 16 * 4, gameFontSmall, "If you clear the whole\nboard you will get a\n1000 point bonus.");
	}
	else
	{
		gfx_font_print(xStart, 20, gameFont, "Controls");
#if COMPILE_TARGET == GEMEI_760P
		gfx_font_print(xStart, 40 + 16 * 0, gameFontSmall, "A to enter menus");
		gfx_font_print(xStart, 40 + 16 * 1, gameFontSmall, "B to go back");
		gfx_font_print(xStart, 40 + 16 * 2, gameFontSmall, "Power/X + Up/Down to\nchange volume");
		gfx_font_print(xStart, 40 + 16 * 5, gameFontSmall, "Ingame:");
		gfx_font_print(xStart, 40 + 16 * 6, gameFontSmall, "A         Remove object");
		gfx_font_print(xStart, 40 + 16 * 7, gameFontSmall, "B         Return to menu");
		gfx_font_print(xStart, 40 + 16 * 8, gameFontSmall, "Triangle  Restart");
#elif COMPILE_TARGET == DINGOO_A320 || COMPILE_TARGET == GEMEI_A330
		gfx_font_print(xStart, 40 + 16 * 0, gameFontSmall, "A to enter menus");
		gfx_font_print(xStart, 40 + 16 * 1, gameFontSmall, "B to go back");
		gfx_font_print(xStart, 40 + 16 * 2, gameFontSmall, "Start + Select anytime\nto quit");
		gfx_font_print(xStart, 40 + 16 * 4, gameFontSmall, "Triggers to change volume");
		gfx_font_print(xStart, 40 + 16 * 6, gameFontSmall, "Ingame:");
		gfx_font_print(xStart, 40 + 16 * 7, gameFontSmall, "A         Remove object");
		gfx_font_print(xStart, 40 + 16 * 8, gameFontSmall, "B/Start   Return to menu");
		gfx_font_print(xStart, 40 + 16 * 9, gameFontSmall, "X         Restart");
#endif
	}

	sprintf(tempString, "Page %u/3", instructionsPage);
	gfx_font_print_center_ex(216, 222, 0, gameFontSmall, tempString);
}

void drawVolumeMeter()
{
	uint16_t w = 80;
	uint16_t h = 20;
	uint16_t screenW = gfx_render_target->width;
	uint16_t screenH = gfx_render_target->height;
	uint16_t yStart = screenH - 40;
	uint16_t xStartLeft = (screenW - w) / 2;

	gfx_rect_fill_draw(xStartLeft, yStart - (h / 2), w, h, gfx_color_rgb(0x60, 0x60, 0x60));
	gfx_rect_draw(xStartLeft, yStart - (h / 2), w, h, gfx_color_rgb(0x40, 0x40, 0x40));
	if (mixerGetVolume() >= 15)
		sprite_draw(volumeHigh, xStartLeft + 2, yStart - 8, 0);
	else if (mixerGetVolume() > 0)
		sprite_draw(volumeLow, xStartLeft + 2, yStart - 8, 0);
	else
		sprite_draw(volumeOff, xStartLeft + 2, yStart - 8, 0);

	gfx_rect_fill_draw(xStartLeft + 20, yStart - (h / 2) + 2, w - 22, h - 4, gfx_color_rgb(0xac, 0xff, 0xa0));
	uint16_t w2 = ((w - 22) * mixerGetVolume()) / VOLUME_MAX;
	gfx_rect_fill_draw(xStartLeft + 20, yStart - (h / 2) + 2, w2, h - 4, gfx_color_rgb(0x18, 0xb3, 0x01));
}

// LOGIC AND LOADING

void doMove(int16_t* x, int16_t* y)
{
	if (inputDelay > 0)
		inputDelay--;

	bool setInputDelay = false;

	if (!control_check(CONTROL_POWER).pressed && !control_check(CONTROL_BUTTON_Y).pressed)
	{
		if ((control_check(CONTROL_DPAD_DOWN).pressed && inputDelay == 0) ||
			control_just_pressed(CONTROL_DPAD_DOWN))
		{
			if (y != NULL)
			{
				(*y)++;
				setInputDelay = true;
			}
		}
		if ((control_check(CONTROL_DPAD_UP).pressed && inputDelay == 0) ||
			control_just_pressed(CONTROL_DPAD_UP))
		{
			if (y != NULL)
			{
				(*y)--;
				setInputDelay = true;
			}
		}
	}
	if ((control_check(CONTROL_DPAD_RIGHT).pressed && inputDelay == 0) ||
		control_just_pressed(CONTROL_DPAD_RIGHT))
	{
		if (x != NULL)
		{
			(*x)++;
			setInputDelay = true;
		}
	}
	if ((control_check(CONTROL_DPAD_LEFT).pressed && inputDelay == 0) ||
		control_just_pressed(CONTROL_DPAD_LEFT))
	{
		if (x != NULL)
		{
			(*x)--;
			setInputDelay = true;
		}
	}

	if (setInputDelay)
		inputDelay = delayAmount;
}

void setGameMode(uint8_t mode)
{
	switch (mode)
	{
		case GAME_MODE_MENU:
			if (gameInProgress)
				menu_index = 0;
			else
				menu_index = 1;

			if (gameMode == GAME_MODE_GAME)
			{
				if (gameModeSub == GAME_MODE_GAME_VIEWHIGHSCORE)
				{
					menu_index = 2;
				}

				setGameSubMode(GAME_MODE_MENU_FADEIN);
			}
			else
			{
				setGameSubMode(GAME_MODE_MENU_MAIN);
				bgRightX = bgRightXMin;
			}
			break;
		case GAME_MODE_GAME:
			setGameSubMode(GAME_MODE_GAME_PLAY);
			break;
	}

	gameMode = mode;
}

void setGameSubMode(uint8_t mode)
{
	if (gameMode == GAME_MODE_GAME)
	{
		switch (mode)
		{
			case GAME_MODE_GAME_ANIMATION:
				disappearingTicks = 0;
				break;
			case GAME_MODE_GAME_WRITEHIGHSCORE:
				if (strlen(inputLastName) == 0)
				{
					markerInputX = 0;
					markerInputY = 0;
				}
				else
				{
					markerInputX = 16;
					markerInputY = 4;
				}
				break;
		}
	}

	if (gameMode == GAME_MODE_MENU)
	{
		if (mode == GAME_MODE_MENU_ACHIEVEMENTS)
		{
			achievementIndex = 0;
		}
		else if (mode == GAME_MODE_MENU_INSTRUCTIONS)
		{
			instructionsPage = 1;
		}
	}

	gameModeSub = mode;
}

uint8_t getTile(int8_t x, int8_t y)
{
	if (x < 0 || y < 0 || x >= GAMEWIDTH || y >= GAMEHEIGHT)
	{
		return 0;
	}

	return grid[x + y * GAMEWIDTH];
}

void setTile(int8_t x, int8_t y, uint8_t tile)
{
	if (x < 0 || y < 0 || x >= GAMEWIDTH || y >= GAMEHEIGHT)
	{
		return;
	}

	grid[x + y * GAMEWIDTH] = tile;
}

void doLogic()
{
	aTimer++;

	// X to restart
	if (control_just_pressed(CONTROL_BUTTON_X))
	{
		startGame();
		return;
	}

	if (control_just_pressed(CONTROL_BUTTON_START) || control_just_pressed(CONTROL_BUTTON_B))
	{
		setGameMode(GAME_MODE_MENU);
		mixerSetSound(sndWoosh, sndWooshSize, false);
		return;
	}

	if (control_just_pressed(CONTROL_BUTTON_A))
	{
		doRemoveTile();
		return;
	}

	int16_t tempX = markerX;
	int16_t tempY = markerY;

	doMove(&markerX, &markerY);

	if (markerX < 0)
		markerX = 0;

	if (markerY < 0)
		markerY = 0;

	if (markerX >= GAMEWIDTH)
		markerX = GAMEWIDTH - 1;

	if (markerY >= GAMEHEIGHT)
		markerY = GAMEHEIGHT - 1;

	// If marker moved
	if (tempX != markerX || tempY != markerY)
	{
		// Remove old animation
		uint8_t tile = getTile(tempX, tempY);
		setTo(TYPE_NORMAL, tile, tempX, tempY);

		// Set new animation
		tile = getTile(markerX, markerY);

		if (getTile(markerX + 1, markerY) == tile ||
			getTile(markerX - 1, markerY) == tile ||
			getTile(markerX, markerY + 1) == tile ||
			getTile(markerX, markerY - 1) == tile)
		{
			setTo(TYPE_SELECTED, tile, markerX, markerY);
		}
	}	
}

void doInput()
{
	doMove(&markerInputX, &markerInputY);
	if (markerInputX < 0)
		markerInputX = 17;
	if (markerInputX >= 18)
		markerInputX = 0;
	if (markerInputY < 0)
		markerInputY = 4;
	if (markerInputY >= 5)
		markerInputY = 0;

	bool end = false;
	bool backspace = false;
	uint8_t length = strlen(inputCurName);

	if (control_just_pressed(CONTROL_BUTTON_A))
	{
		if (markerInputY <= 3)
		{
			if (length < 14)
			{
				inputCurName[length] = inputChars[markerInputX + markerInputY * 18];
				inputCurName[length + 1] = '\0';
			}
		}
		else
		{
			if (markerInputX >= 4 && markerInputX <= 10)
			{
				backspace = true;
			}
			if (markerInputX >= 15 && markerInputX <= 17)
			{
				end = true;
			}
		}
	}

	if (control_just_pressed(CONTROL_BUTTON_START))
	{
		end = true;
	}

	if (backspace)
	{
		if (length > 0)
			inputCurName[length - 1] = '\0';
	}

	if (end)
	{
		if (length > 0)
		{
			sprintf(inputLastName, "%s", inputCurName);

			// Loop through and find our place in the score board
			uint8_t i;
			for (i = 0; i < 6; i++)
			{
				if (score > highscores[i].score)
				{
					// Move down other scores
					uint8_t j;
					for (j = 5; j > i; j--)
					{
						highscores[j].score = highscores[j - 1].score;
						sprintf(highscores[j].name, "%s", highscores[j - 1].name);
					}

					highscores[i].score = score;
					sprintf(highscores[i].name, "%s", inputCurName);

					break;
				}
			}
			setGameSubMode(GAME_MODE_GAME_VIEWHIGHSCORE);
		}
	}
}

void doRemoveTile()
{
	uint8_t tile = getTile(markerX, markerY);
	
	if (tile == TILE_NONE)
	{
		return;
	}

	if (getTile(markerX + 1, markerY) == tile ||
		getTile(markerX - 1, markerY) == tile ||
		getTile(markerX, markerY + 1) == tile ||
		getTile(markerX, markerY - 1) == tile)
	{
		setGameSubMode(GAME_MODE_GAME_ANIMATION);
		uint16_t n = setTo(TYPE_DISAPPEARING, tile, markerX, markerY);
		uint32_t newScore = pow(n - 2, 2);
		score += newScore;
		if (newScore >= 2000)
			mixerSetSound(sndPopBig, sndPopBigSize, false);
		else
			mixerSetSound(sndPop, sndPopSize, false);
	}
}

uint8_t getNormalColorTile(uint8_t tile)
{
	if (tile == TILE_RED || tile == TILE_RED_SELECTED || tile == TILE_RED_DISAPPEARING)
		return TILE_RED;
	if (tile == TILE_BLUE || tile == TILE_BLUE_SELECTED || tile == TILE_BLUE_DISAPPEARING)
		return TILE_BLUE;
	if (tile == TILE_YELLOW || tile == TILE_YELLOW_SELECTED || tile == TILE_YELLOW_DISAPPEARING)
		return TILE_YELLOW;

	return TILE_NONE;
}

uint8_t getTileType(uint8_t tile)
{
	if (tile >= TILE_RED && tile <= TILE_YELLOW)
		return TYPE_NORMAL;
	if (tile >= TILE_RED_SELECTED && tile <= TILE_YELLOW_SELECTED)
		return TYPE_SELECTED;
	if (tile >= TILE_RED_DISAPPEARING && tile <= TILE_YELLOW_DISAPPEARING)
		return TYPE_DISAPPEARING;

	return TYPE_NONE;
}

uint16_t setTo(uint8_t type, uint8_t tileToSet, int8_t x, int8_t y)
{
	uint8_t tile = getTile(x, y);
	if (tile != tileToSet)
		return 0;

	if (getTileType(tile) == type)
		return 0;

	uint8_t normalTile = getNormalColorTile(tile);

	if (normalTile == TILE_NONE)
		return 0;

	uint16_t n = 1;

	switch (normalTile)
	{
		case TILE_RED:
			if (type == TYPE_NORMAL)
				setTile(x, y, TILE_RED);
			else if (type == TYPE_DISAPPEARING)
				setTile(x, y, TILE_RED_DISAPPEARING);
			else if (type == TYPE_SELECTED)
				setTile(x, y, TILE_RED_SELECTED);
			break;
		case TILE_BLUE:
			if (type == TYPE_NORMAL)
				setTile(x, y, TILE_BLUE);
			else if (type == TYPE_DISAPPEARING)
				setTile(x, y, TILE_BLUE_DISAPPEARING);
			else if (type == TYPE_SELECTED)
				setTile(x, y, TILE_BLUE_SELECTED);
			break;
		default:
			if (type == TYPE_NORMAL)
				setTile(x, y, TILE_YELLOW);
			else if (type == TYPE_DISAPPEARING)
				setTile(x, y, TILE_YELLOW_DISAPPEARING);
			else if (type == TYPE_SELECTED)
				setTile(x, y, TILE_YELLOW_SELECTED);
			break;
	}
	
	n += setTo(type, tileToSet, x + 1, y);
	n += setTo(type, tileToSet, x - 1, y);
	n += setTo(type, tileToSet, x, y + 1);
	n += setTo(type, tileToSet, x, y - 1);

	return n;
}

void doDisappearing()
{
	aTimer++;
	disappearingTicks++;
	if (disappearingTicks >= 11)
	{
		removeAllDisappearing();
		if (isGameOver())
		{
			gameInProgress = false;

			if (score >= 3000)
				unlockAchievement(ACHIEVEMENT_HIGHSCORE1);

			if (score >= 4000)
				unlockAchievement(ACHIEVEMENT_HIGHSCORE2);

			cumulativeScore += score;

			if (cumulativeScore >= 75000)
				unlockAchievement(ACHIEVEMENT_CUMULATIVE1);

			if (cumulativeScore >= 250000)
				unlockAchievement(ACHIEVEMENT_CUMULATIVE2);

			if (aTimer <= 30 * 30 && score >= 2000)
				unlockAchievement(ACHIEVEMENT_SPEEDY);

			if (score > highscores[5].score)
			{
				mixerSetSound(sndHighscore, sndHighscoreSize, false);
				setGameSubMode(GAME_MODE_GAME_WRITEHIGHSCORE);
			}
			else
			{
				setGameSubMode(GAME_MODE_GAME_VIEWHIGHSCORE);
			}
		}
		else
		{
			setGameSubMode(GAME_MODE_GAME_PLAY);
		}
	}
}

void removeAllDisappearing()
{
	int16_t x;
	int16_t y;

	// Achievement stuff
	bool foundRed = false;
	bool foundBlue = false;
	bool foundYellow = false;
	uint8_t curRemove = TILE_NONE;

	// Set to NONE
	for (y = 0; y < GAMEHEIGHT; y++)
	{
		for (x = 0; x < GAMEWIDTH; x++)
		{
			uint8_t tile = getTile(x, y);

			if (tile >= TILE_RED_DISAPPEARING)
			{
				curRemove = tile - 6;
				setTile(x, y, TILE_NONE);
			}
			else if (tile == TILE_RED)
			{
				foundRed = true;
			}
			else if (tile == TILE_BLUE)
			{
				foundBlue = true;
			}
			else if (tile == TILE_YELLOW)
			{
				foundYellow = true;
			}
		}
	}

	// Achievement stuff
	if (curRemove == TILE_RED)
	{
		if (!removedAnyRed && !foundRed)
		{
			unlockAchievement(ACHIEVEMENT_EXTINCTION);
		}

		removedAnyRed = true;
	}
	else if (curRemove == TILE_BLUE)
	{
		if (!removedAnyBlue && !foundBlue)
		{
			unlockAchievement(ACHIEVEMENT_EXTINCTION);
		}

		removedAnyBlue = true;
	}
	else if (curRemove == TILE_YELLOW)
	{
		if (!removedAnyYellow && !foundYellow)
		{
			unlockAchievement(ACHIEVEMENT_EXTINCTION);
		}

		removedAnyYellow = true;
	}

	// Drop down tiles
	for (y = GAMEHEIGHT - 1; y >= 0; y--)
	{
		for (x = GAMEWIDTH - 1; x >= 0; x--)
		{
			if (getTile(x, y) == TILE_NONE)
			{
				int16_t y2;
				for (y2 = y - 1; y2 >= 0; y2--)
				{
					uint8_t tempTile = getTile(x, y2);
					if (tempTile != TILE_NONE)
					{
						setTile(x, y, tempTile);
						setTile(x, y2, TILE_NONE);
						break;
					}
				}
			}
		}
	}

	// Move to left if empty columns
	for (x = GAMEWIDTH - 1; x >= 0; x--)
	{
		y = GAMEHEIGHT - 1;
		if (getTile(x, y) == TILE_NONE)
		{
			int16_t x2;
			int16_t y2;

			// Move everything that is right of the empty column left
			for (x2 = x; x2 < GAMEWIDTH; x2++)
			{
				for (y2 = y; y2 >= 0; y2--)
				{
					uint8_t tempTile = getTile(x2 + 1, y2);
					setTile(x2, y2, tempTile);
				}
			}
		}
	}

	// Set selected tiles
	uint8_t tile = getTile(markerX, markerY);

	if (getTile(markerX + 1, markerY) == tile ||
		getTile(markerX - 1, markerY) == tile ||
		getTile(markerX, markerY + 1) == tile ||
		getTile(markerX, markerY - 1) == tile)
	{
		setTo(TYPE_SELECTED, tile, markerX, markerY);
	}
}

bool isGameOver()
{
	// Check if game is over
	bool gameOver = true;
	int16_t x;
	int16_t y;
	uint8_t tile;
	uint16_t tilesLeft = 0;

	for (y = 0; y < GAMEHEIGHT; y++)
	{
		for (x = 0; x < GAMEWIDTH; x++)
		{
			tile = getTile(x, y);
			if (tile != TILE_NONE)
			{
				tilesLeft++;

				if (getTile(x + 1, y) == tile ||
					getTile(x - 1, y) == tile ||
					getTile(x, y + 1) == tile ||
					getTile(x, y - 1) == tile)
				{
					gameOver = false;
				}
			}
		}
	}

	if (tilesLeft == 0)
	{
		unlockAchievement(ACHIEVEMENT_CLEAR);
		score += 1000;
	}

	return gameOver;
}

void createLevel()
{
	int16_t  x;
	int16_t  y;
	uint8_t  lastColor;
	uint16_t lastColorTimes;
	uint8_t  r;
	uint8_t  red;
	uint8_t  blue;
	uint8_t  yellow;

	do
	{
		lastColor = 0;
		lastColorTimes = 0;
		red = 0;
		blue = 0;
		yellow = 0;

		for (y = 0; y < GAMEHEIGHT; y++)
		{
			for (x = 0; x < GAMEWIDTH; x++)
			{
				do
				{
					r = (rand() % 3) + 1;
				} while (r == lastColor && lastColorTimes > 2);

				if (r == lastColor)
				{
					lastColorTimes++;
				}
				else
				{
					lastColor = r;
					lastColorTimes = 1;
				}

				switch (r)
				{
					case TILE_RED:
						red++;
						break;
					case TILE_BLUE:
						blue++;
						break;
					case TILE_YELLOW:
						yellow++;
						break;
				}

				grid[x + y * GAMEWIDTH] = r;
			}
		}
	} while (red > 55 || blue > 55 || yellow > 55);
}

void startGame()
{
	aTimer = 0;
	removedAnyRed = false;
	removedAnyBlue = false;
	removedAnyYellow = false;
	score = 0;
	gameInProgress = true;
	sprintf(inputCurName, "%s", inputLastName);
	createLevel();
	startAnimation = 0;
	
	// Set selected tiles
	uint8_t tile = getTile(markerX, markerY);

	if (getTile(markerX + 1, markerY) == tile ||
		getTile(markerX - 1, markerY) == tile ||
		getTile(markerX, markerY + 1) == tile ||
		getTile(markerX, markerY - 1) == tile)
	{
		setTo(TYPE_SELECTED, tile, markerX, markerY);
	}
}

bool loadGame()
{
	gameFont = gfx_font_load_hif_from_buffer(hifFont);
	gameFontDisabled = gfx_font_load_hif_from_buffer(hifFontDisabled);
	gameFontSmall = gfx_font_load_hif_from_buffer(hifFontSmall);
	gameFontAchievement = gfx_font_load_hif_from_buffer(hifFontAchievement);

	gameLogo = gfx_tex_load_hif_from_buffer(hifGameLogo);
	gameBackground = gfx_tex_load_hif_from_buffer(hifBackground);
	gameBackgroundRight = gfx_tex_load_hif_from_buffer(hifBackgroundRight);

	tileRed = sprite_load_from_hif_buffer(hifRed, COLOR_MAGENTA);
	tileBlue = sprite_load_from_hif_buffer(hifBlue, COLOR_MAGENTA);
	tileYellow = sprite_load_from_hif_buffer(hifYellow, COLOR_MAGENTA);
	gameMarker = sprite_load_from_hif_buffer(hifMarker, COLOR_MAGENTA);
	menuArrowLeft = sprite_load_from_hif_buffer(hifMenuArrowLeft, COLOR_MAGENTA);
	menuArrowRight = sprite_load_from_hif_buffer(hifMenuArrowRight, COLOR_MAGENTA);
	achievementLocked = sprite_load_from_hif_buffer(hifAchievementLocked, COLOR_MAGENTA);
	achievementUnlocked = sprite_load_from_hif_buffer(hifAchievementUnlocked, COLOR_MAGENTA);
	achievementCheck = sprite_load_from_hif_buffer(hifAchievementCheck, COLOR_MAGENTA);
	volumeHigh = sprite_load_from_hif_buffer(hifVolumeHigh, COLOR_MAGENTA);
	volumeLow = sprite_load_from_hif_buffer(hifVolumeLow, COLOR_MAGENTA);
	volumeOff = sprite_load_from_hif_buffer(hifVolumeOff, COLOR_MAGENTA);
	tileRedSelected = sprite_load_from_hif_buffer(hifRedSelected, COLOR_MAGENTA);
	tileBlueSelected = sprite_load_from_hif_buffer(hifBlueSelected, COLOR_MAGENTA);
	tileYellowSelected = sprite_load_from_hif_buffer(hifYellowSelected, COLOR_MAGENTA);

	if (gameFont == NULL || gameLogo == NULL || tileRed == NULL ||
		tileBlue == NULL || tileYellow == NULL || gameBackground == NULL ||
		gameBackgroundRight == NULL || gameMarker == NULL || menuArrowLeft == NULL ||
		menuArrowRight == NULL || gameFontDisabled == NULL || gameFontAchievement == NULL ||
		achievementLocked == NULL || achievementUnlocked == NULL || achievementCheck == NULL ||
		gameFontSmall == NULL || harteexLogo == NULL || volumeHigh == NULL ||
		volumeLow == NULL || volumeOff == NULL || tileRedSelected == NULL || 
		tileBlueSelected == NULL || tileYellowSelected == NULL)
	{
		errorMsg = "Unable to load graphics (1)";
		return false;
	}

	tileRedAnim = sprite_create(tileRed->width, tileRed->height);
	tileBlueAnim = sprite_create(tileBlue->width, tileBlue->height);
	tileYellowAnim = sprite_create(tileYellow->width, tileYellow->height);

	if (tileRedAnim == NULL || tileBlueAnim == NULL || tileYellowAnim == NULL)
	{
		errorMsg = "Unable to load graphics (2)";
		return false;
	}

	uint16_t i;
	uint16_t tileW = tileRed->width; // Let's assume all are of the same size
	uint16_t tileH = tileRed->height;
	gfx_texture* tempSrcTexRed = gfx_tex_load_hif_from_buffer(hifRedSelected);
	gfx_texture* tempSrcTexBlue = gfx_tex_load_hif_from_buffer(hifBlueSelected);
	gfx_texture* tempSrcTexYellow = gfx_tex_load_hif_from_buffer(hifYellowSelected);
	gfx_texture* tempTrgTex = gfx_tex_allocate(tileW, tileH);
	if (tempSrcTexRed == NULL || tempSrcTexBlue == NULL || tempSrcTexYellow == NULL || tempTrgTex == NULL)
	{
		errorMsg = "Unable to load graphics (3)";
		return false;
	}
#ifdef SML_COLOR_32BIT
	uint32_t* tempSrcTexPtr;
	uint32_t* tempTrgTexPtr = tempTrgTex->address;
#else
	uint16_t* tempSrcTexPtr;
	uint16_t* tempTrgTexPtr = tempTrgTex->address;
#endif
	uint8_t tempRed;
	uint8_t tempGreen;
	uint8_t tempBlue;
	uint8_t switcher;
	for (switcher = 0; switcher < 3; switcher++)
	{
		switch (switcher)
		{
			case 0:
				tempSrcTexPtr = tempSrcTexRed->address;
				break;
			case 1:
				tempSrcTexPtr = tempSrcTexBlue->address;
				break;
			default:
				tempSrcTexPtr = tempSrcTexYellow->address;
				break;
		}

		for (i = 0; i < tileAnimFrames; i++)
		{
			uint32_t j;
			for (j = 0; j < tileW * tileH; j++)
			{				
				if (tempSrcTexPtr[j] == COLOR_MAGENTA)
				{
					tempTrgTexPtr[j] = COLOR_MAGENTA;
					continue;
				}
#ifdef SML_COLOR_32BIT
				uint32_t color = tempSrcTexPtr[j];
#else
				uint32_t color = gfx_color_back(tempSrcTexPtr[j]);
#endif
				tempRed = (uint8_t)(color >> 16);
				tempGreen = (uint8_t)(color >> 8);
				tempBlue = (uint8_t)color;
				
				// Move towards the color 0x76afdc

				int16_t diffR = 0x76 - tempRed;
				int16_t diffG = 0xaf - tempGreen;
				int16_t diffB = 0xdc - tempBlue;

				int16_t addR = ((i + 1) * diffR) / tileAnimFrames;
				int16_t addG = ((i + 1) * diffG) / tileAnimFrames;
				int16_t addB = ((i + 1) * diffB) / tileAnimFrames;

				tempTrgTexPtr[j] = gfx_color_rgb(tempRed + addR, tempGreen + addG, tempBlue + addB);
			}

			switch (switcher)
			{
				case 0:
					tileRedAnim = sprite_frame_add_bitmap(tileRedAnim, tempTrgTexPtr, COLOR_MAGENTA);
					break;
				case 1:
					tileBlueAnim = sprite_frame_add_bitmap(tileBlueAnim, tempTrgTexPtr, COLOR_MAGENTA);
					break;
				default:
					tileYellowAnim = sprite_frame_add_bitmap(tileYellowAnim, tempTrgTexPtr, COLOR_MAGENTA);
					break;
			}
		}
	}

	gfx_tex_delete(tempSrcTexRed);
	gfx_tex_delete(tempSrcTexBlue);
	gfx_tex_delete(tempSrcTexYellow);
	gfx_tex_delete(tempTrgTex);

	return true;
}

void unloadGame()
{
	gfx_font_delete(gameFont);
	gfx_font_delete(gameFontDisabled);
	gfx_font_delete(gameFontSmall);
	gfx_font_delete(gameFontAchievement);

	gfx_tex_delete(gameLogo);
	gfx_tex_delete(harteexLogo);
	gfx_tex_delete(gameBackground);
	gfx_tex_delete(gameBackgroundRight);

	sprite_delete(tileRed);
	sprite_delete(tileBlue);
	sprite_delete(tileYellow);
	sprite_delete(gameMarker);
	sprite_delete(menuArrowLeft);
	sprite_delete(menuArrowRight);
	sprite_delete(achievementLocked);
	sprite_delete(achievementUnlocked);
	sprite_delete(achievementCheck);
	sprite_delete(volumeHigh);
	sprite_delete(volumeLow);
	sprite_delete(volumeOff);
	sprite_delete(tileRedAnim);
	sprite_delete(tileBlueAnim);
	sprite_delete(tileYellowAnim);
	sprite_delete(tileRedSelected);
	sprite_delete(tileBlueSelected);
	sprite_delete(tileYellowSelected);
}

bool loadData()
{
	bool error = false;
	
	FILE* tempFile = fopen("SameGoo.dat", "rb");
	if (tempFile == NULL)
	{
		error = true;
	}
	else
	{
		// Check file
		char id[3];

		fread(&id[0], 1, 1, tempFile);
		fread(&id[1], 1, 1, tempFile);
		fread(&id[2], 1, 1, tempFile);

		if (strncmp(id, "SGD", 3) != 0)
		{
			fclose(tempFile);
			error = true;
		}
	}

	if (error)
	{
		achievementsTaken = 0;
		cumulativeScore = 0;
		gameStarts = 0;

		uint16_t i;
		for (i = 0; i < 6; i++)
		{
			sprintf(highscores[i].name, "None");
			highscores[i].score = 0;
		}

		inputLastName[0] = '\0';

		return false;
	}

	// Read the real data

	uint16_t i;
	for (i = 0; i < 6; i++)
	{
		fread(&highscores[i].name, 15, 1, tempFile);
		fread(&highscores[i].score, 4, 1, tempFile);
	}

	fread(&achievementsTaken, 1, 1, tempFile);
	fread(inputLastName, 15, 1, tempFile);
	fread(&cumulativeScore, 4, 1, tempFile);
	fread(&gameStarts, 2, 1, tempFile);

	if (!feof(tempFile)) // Needed for compatibility with SameGoo 1.0 dat-file
	{
		fread(&volumeConfig, 1, 1, tempFile);
	}

	fclose(tempFile);
	return true;
}

bool saveData()
{
	FILE* tempFile = fopen("SameGoo.dat", "wb");
	if (tempFile == NULL)
		return false;

	char* id = "SGD";
	fwrite(&id[0], 1, 1, tempFile);
	fwrite(&id[1], 1, 1, tempFile);
	fwrite(&id[2], 1, 1, tempFile);

	uint16_t i;
	for (i = 0; i < 6; i++)
	{
		fwrite(&highscores[i].name, 15, 1, tempFile);
		fwrite(&highscores[i].score, 4, 1, tempFile);
	}

	fwrite(&achievementsTaken, 1, 1, tempFile);
	fwrite(inputLastName, 15, 1, tempFile);
	fwrite(&cumulativeScore, 4, 1, tempFile);
	fwrite(&gameStarts, 2, 1, tempFile);

	volumeConfig = mixerGetVolume();
	fwrite(&volumeConfig, 1, 1, tempFile);

	fclose(tempFile);
	return true;
}

uint8_t _rand8()
{
	uint32_t tempRand = rand();
	return ((tempRand >> 24) ^ (tempRand >> 16) ^ (tempRand >> 8) ^ tempRand);
}

int main(int argc, char** argv) {
	int ref = EXIT_SUCCESS;

	srand(OSTimeGet());

	control_init();
	gameDisplay = display_create(320, 240, 320, (DISPLAY_FORMAT_RGB565 | DISPLAY_BUFFER_STATIC), NULL, NULL);
	if(gameDisplay == NULL) {
		control_term();
		return ref;
	}
	gfx_init(gameDisplay);

	harteexLogo = gfx_tex_load_hif_from_buffer(hifLogo);
	drawSplash();
	display_flip(gameDisplay);

	control_lock(timer_resolution / 4);

	gameTimer   = timer_create();

	uint32_t tempTick  = 0;
	uint32_t tempPauseOS = 0;
	
	uint32_t splashTicks = 0;

	if (!loadGame())
		gameMode = GAME_MODE_ERROR;

	curAchievements.count = 0;
	curAchievements.first = NULL;

	loadData();
	gameStarts++;
	if (gameStarts >= 20)
	{
		unlockAchievement(ACHIEVEMENT_ADDICTED);
	}

	//sprintf(tempString, "%sSameGoo.ogg", gamePath);
	mixerInit(44100, 16, 2, volumeConfig);
	if (!mixerSetOgg("SameGoo.ogg"))
	{
		mixerSetOggFromBuffer(oggMusic, oggMusicSize);
	}
	mixerPause(false);

	int sysref;
	while(gameRunning) {
#if COMPILE_TARGET == GEMEI_A330
		sysref = sys_judge_event(NULL);
#else
		sysref = _sys_judge_event(NULL);
#endif
		if(sysref < 0) {
			ref = sysref;
			break;
		}

		tempTick += timer_delta(gameTimer);
		if(tempTick < gameTickRate) {
			tempPauseOS = (((gameTickRate - tempTick) * OS_TICKS_PER_SEC) / timer_resolution);
			if(tempPauseOS > 0) OSTimeDly(tempPauseOS);
			while(tempTick < gameTickRate)
				tempTick += timer_delta(gameTimer);
		}

		// --- UPDATE ---
		mixerUpdate();

		while(tempTick >= gameTickRate) {
			tempTick -= gameTickRate;

			control_poll();
			if (control_check(CONTROL_BUTTON_START).pressed && control_check(CONTROL_BUTTON_SELECT).pressed)
			{
				gameRunning = false;
				break;
			}

			if (control_check(CONTROL_BUTTON_SELECT).pressed)
			{
				screenshotCounter++;
			}
			else
			{
				screenshotCounter = 0;
			}

			if (screenshotCounter == 90)
				gameScreenshot();

			if (checkAchievements())
			{
				mixerSetSound(sndAchievement, sndAchievementSize, false);
			}

			if (showVolumeMeter > 0)
				showVolumeMeter--;

			if (control_just_pressed(CONTROL_TRIGGER_LEFT) ||
				((control_check(CONTROL_POWER).pressed || control_check(CONTROL_BUTTON_Y).pressed) &&
				control_just_pressed(CONTROL_DPAD_DOWN)))
			{
				mixerVolumeDecrease();
				showVolumeMeter = 50;
			}

			if (control_just_pressed(CONTROL_TRIGGER_RIGHT) ||
				((control_check(CONTROL_POWER).pressed || control_check(CONTROL_BUTTON_Y).pressed) &&
				control_just_pressed(CONTROL_DPAD_UP)))
			{
				mixerVolumeIncrease();
				showVolumeMeter = 50;
			}

			switch (gameMode) {
				case GAME_MODE_SPLASH:
					splashTicks++;
					if (splashTicks >= 100 || control_just_pressed(CONTROL_BUTTON_A) || control_just_pressed(CONTROL_BUTTON_START))
					{
						setGameMode(GAME_MODE_MENU);
					}
					break;
				case GAME_MODE_MENU:
					switch (gameModeSub)
					{
						case GAME_MODE_MENU_MAIN:
							if (control_just_pressed(CONTROL_DPAD_UP) && !control_check(CONTROL_POWER).pressed &&
								!control_check(CONTROL_BUTTON_Y).pressed)
							{
								if (menu_index > 1 || (menu_index > 0 && gameInProgress))
								{
									menu_index--;
								}
								else
								{
									menu_index = menu_length - 1;
								}
							}
							if (control_just_pressed(CONTROL_DPAD_DOWN) && !control_check(CONTROL_POWER).pressed &&
								!control_check(CONTROL_BUTTON_Y).pressed)
							{
								if (menu_index < menu_length - 1)
								{
									menu_index++;
								}
								else
								{
									if (!gameInProgress)
									{
										menu_index = 1;
									}
									else
									{
										menu_index = 0;
									}
								}
							}

							if (control_just_pressed(CONTROL_BUTTON_A))
							{
								switch (menu_index) {
									case 0:
										// CONTINUE
										mixerSetSound(sndWoosh, sndWooshSize, false);
										setGameSubMode(GAME_MODE_MENU_FADEOUT);
										targetSub = GAME_MODE_GAME_PLAY;
										break;
									case 1:
										// START NEW GAME
										mixerSetSound(sndWoosh, sndWooshSize, false);
										setGameSubMode(GAME_MODE_MENU_FADEOUT);
										targetSub = GAME_MODE_GAME_PLAY;
										startGame();
										break;
									case 2:
										// HIGH SCORES
										mixerSetSound(sndWoosh, sndWooshSize, false);
										setGameSubMode(GAME_MODE_MENU_FADEOUT);
										targetSub = GAME_MODE_GAME_VIEWHIGHSCORE;
										break;
									case 3:
										// INSTRUCTIONS
										setGameSubMode(GAME_MODE_MENU_INSTRUCTIONS);
										break;
									case 4:
										// ACHIEVEMENTS
										setGameSubMode(GAME_MODE_MENU_ACHIEVEMENTS);
										break;
									case 5:
										// CREDITS
										setGameSubMode(GAME_MODE_MENU_CREDITS);
										break;
									case 6:
										gameRunning = false;
										break;
								}
							}
							break;
						case GAME_MODE_MENU_FADEIN:
							bgRightX -= 8;

							if (bgRightX <= bgRightXMin)
							{
								bgRightX = bgRightXMin;
								setGameSubMode(GAME_MODE_MENU_MAIN);
							}
							break;
						case GAME_MODE_MENU_FADEOUT:
							bgRightX += 8;

							if (bgRightX >= bgRightXMax)
							{
								bgRightX = bgRightXMax;
								setGameMode(GAME_MODE_GAME);
								setGameSubMode(targetSub);
							}
							break;
						case GAME_MODE_MENU_ACHIEVEMENTS:
							doMove(NULL, &achievementIndex);

							if (achievementIndex < 0)
								achievementIndex = 7;
							if (achievementIndex >= 8)
								achievementIndex = 0;

							if (control_just_pressed(CONTROL_BUTTON_B))
							{
								setGameSubMode(GAME_MODE_MENU_MAIN);
							}
							break;
						case GAME_MODE_MENU_CREDITS:
							if (control_just_pressed(CONTROL_BUTTON_B))
							{
								setGameSubMode(GAME_MODE_MENU_MAIN);
							}
							break;

						case GAME_MODE_MENU_INSTRUCTIONS:
							if (control_just_pressed(CONTROL_BUTTON_B))
							{
								setGameSubMode(GAME_MODE_MENU_MAIN);
							}

							if (!control_check(CONTROL_POWER).pressed && !control_check(CONTROL_BUTTON_Y).pressed)
							{
								if ((instructionsPage == 1 || instructionsPage == 2) &&
									(control_just_pressed(CONTROL_DPAD_RIGHT) || control_just_pressed(CONTROL_DPAD_DOWN)))
								{
									instructionsPage++;
								}

								if ((instructionsPage == 2 || instructionsPage == 3) &&
									(control_just_pressed(CONTROL_DPAD_LEFT) || control_just_pressed(CONTROL_DPAD_UP)))
								{
									instructionsPage--;
								}
							}
							break;
					}
					break;
				case GAME_MODE_GAME:
					if (startAnimation < TILESIZE / 2)
					{
						startAnimation++;
					}

					switch(gameModeSub)
					{
						case GAME_MODE_GAME_PLAY:
							doLogic();
							break;
						case GAME_MODE_GAME_ANIMATION:
							doDisappearing();
							break;
						case GAME_MODE_GAME_WRITEHIGHSCORE:
							doInput();
							break;
						case GAME_MODE_GAME_VIEWHIGHSCORE:
							if (control_just_pressed(CONTROL_BUTTON_A))
							{
								setGameSubMode(GAME_MODE_GAME_PLAY);
								if (!gameInProgress)
									startGame();
							}
							else if (control_just_pressed(CONTROL_BUTTON_B))
							{
								mixerSetSound(sndWoosh, sndWooshSize, false);
								setGameMode(GAME_MODE_MENU);
							}
							break;
						case GAME_MODE_GAME_END:
							break;
					}
					break;
				case GAME_MODE_ERROR:
					if (control_just_pressed(CONTROL_BUTTON_START) || 
						control_just_pressed(CONTROL_BUTTON_A) || 
						control_just_pressed(CONTROL_BUTTON_B))
					{
						gameRunning = false;
					}
					break;
			}
		}
		// --- DRAW ---
		switch(gameMode) {
			case GAME_MODE_SPLASH:
				drawSplash();
				break;
			case GAME_MODE_MENU:
				drawMenu();
				break;
			case GAME_MODE_GAME:
				drawGame();
				break;
			case GAME_MODE_ERROR:
				drawError();
				break;
		}

		drawAchievementPopup(220, 40, gfx_render_target->width, gfx_render_target->height);

		if (showVolumeMeter > 0)
		{
			drawVolumeMeter();
		}

		display_flip(gameDisplay);
	}

	mixerPause(true);

	bgRightX = bgRightXMin;
	gfx_tex_draw(0, 0, gameBackground);
	gfx_tex_draw(bgRightX, 0, gameBackgroundRight);
	gfx_font_print_center_ex(180, 222, 0, gameFont, "Saving...");
	display_flip(gameDisplay);

	saveData();

	gfx_tex_draw(0, 0, gameBackground);
	gfx_tex_draw(bgRightX, 0, gameBackgroundRight);
	gfx_font_print_center_ex(180, 222, 0, gameFont, "Exiting...");
	display_flip(gameDisplay);

	mixerClose();

	timer_delete(gameTimer);

	unloadGame();

	gfx_term();
	display_delete(gameDisplay);
	control_term();

	return ref;
}
